import sys
import os
import torch
import fire
from opendr.engine.datasets import PointCloudsDatasetIterator
from opendr.perception.object_detection_3d import VoxelObjectDetection3DLearner
from opendr.perception.object_detection_3d import (
    KittiDataset,
    LabeledPointCloudsDatasetIterator,
)

config_roots = {
    "pointpillars": os.path.join(
        ".",
        "src",
        "opendr",
        "perception",
        "object_detection_3d",
        "voxel_object_detection_3d",
        "second_detector",
        "configs",
        "pointpillars",
        "car",
    ),
    "tanet": os.path.join(
        ".",
        "src",
        "opendr",
        "perception",
        "object_detection_3d",
        "voxel_object_detection_3d",
        "second_detector",
        "configs",
        "tanet",
        "car",
    ),
}
temp_dir = "./run/models"
subsets_path = os.path.join(
    ".",
    "src",
    "opendr",
    "perception",
    "object_detection_3d",
    "datasets",
    "kitti_subsets",
)
dataset = KittiDataset("/data/sets/opendr_kitti/", subsets_path)


def save_model(
    model_type,
    device="cuda:0",
    load=0,
    name="pointpillars_car",
    config="xyres_16.proto",
    num_ensembles=None,
):

    config = os.path.join(config_roots[model_type], config,)
    model_path = os.path.join(temp_dir, name)

    learner = VoxelObjectDetection3DLearner(
        model_config_path=config,
        device=device,
        checkpoint_after_iter=1000,
        checkpoint_load_iter=load,
        return_uncertainty=True,
        num_ensembles=num_ensembles,
    )

    learner.save(model_path)


def train_model(
    model_type,
    device="cuda:0",
    load=0,
    name="pointpillars_car",
    config="xyres_16.proto",
    samples_list=None,
):

    config = os.path.join(config_roots[model_type], config,)
    model_path = os.path.join(temp_dir, name)

    append_samples_to_name = samples_list is not None

    if samples_list is None:
        samples_list = [1]

    for samples in samples_list:
        if append_samples_to_name:
            model_path = os.path.join(temp_dir, name + "_s" + str(samples))

        learner = VoxelObjectDetection3DLearner(
            model_config_path=config,
            device=device,
            checkpoint_after_iter=1000,
            checkpoint_load_iter=load,
            return_uncertainty=True,
            num_ensembles=samples,
        )

        learner.fit(
            dataset, model_dir=model_path, verbose=False, evaluate=False, samples=samples
        )
        learner.save(model_path)


def train_pointpillars(
    device="cuda:0",
    load=0,
    name="pointpillars_car",
    config="xyres_16.proto",
    samples_list=None,
):
    return train_model(
        "pointpillars",
        device=device,
        load=load,
        name=name,
        samples_list=samples_list,
        config=config,
    )


def train_tanet(
    device="cuda:0",
    load=0,
    name="tanet_car",
    config="xyres_16.proto",
    samples_list=None,
):
    return train_model(
        "tanet",
        device=device,
        load=load,
        name=name,
        samples_list=samples_list,
        config=config,
    )


def train_tapp(
    device="cuda:0",
    load=0,
    name="tapp_car",
    config="tapp_16.proto",
    samples_list=None,
):
    return train_model(
        "tanet",
        device=device,
        load=load,
        name=name,
        samples_list=samples_list,
        config=config,
    )


def train_vnn_pointpillars(
    device="cuda:0",
    load=0,
    name="vnn_pointpillars_car",
    samples_list=[1, 2, 3, 4],
    config="vnn_xyres_16.proto",
):
    return train_pointpillars(
        device=device, load=load, name=name, samples_list=samples_list, config=config
    )


def train_vnna_pointpillars(
    device="cuda:0",
    load=0,
    name="vnna_pointpillars_car",
    samples_list=[1, 2, 3, 4],
    config="vnna_xyres_16.proto",
):
    return train_pointpillars(
        device=device, load=load, name=name, samples_list=samples_list, config=config
    )



def train_lens_pointpillars(
    device="cuda:0",
    load=0,
    name="lens_pointpillars_car",
    samples_list=[2],
    config="lens_xyres_16.proto",
):
    return train_pointpillars(
        device=device, load=load, name=name, samples_list=samples_list, config=config
    )

def train_vnn_tanet(
    device="cuda:0",
    load=0,
    name="mvnn_tanet_car",
    config="vnn_xyres_16.proto",
    samples_list=[5, 1],
):
    return train_model(
        "tanet",
        device=device,
        load=load,
        name=name,
        samples_list=samples_list,
        config=config,
    )


def train_ens_tanet(
    device="cuda:0",
    load=0,
    name="ens_tanet_car",
    config="ens_xyres_16.proto",
    samples_list=[4, 3, 2],
):
    return train_model(
        "tanet",
        device=device,
        load=load,
        name=name,
        samples_list=samples_list,
        config=config,
    )



def train_lens_tanet(
    device="cuda:0",
    load=0,
    name="lens_tanet_car",
    config="lens_xyres_16.proto",
    samples_list=[4, 3, 2],
):
    return train_model(
        "tanet",
        device=device,
        load=load,
        name=name,
        samples_list=samples_list,
        config=config,
    )

def train_vnn_tapp(
    device="cuda:0",
    load=0,
    name="vnn_tapp_car",
    config="vnn_tapp_16.proto",
    samples_list=[2, 1],
):
    return train_model(
        "tanet",
        device=device,
        load=load,
        name=name,
        samples_list=samples_list,
        config=config,
    )


def test_model(
    model_type,
    device="cuda:0",
    name="pointpillars_car",
    samples_list=[1],
    config="xyres_16.proto",
    eval_suffix="classic",
):

    config = os.path.join(config_roots[model_type], config,)
    model_path = os.path.join(temp_dir, name)

    learner = VoxelObjectDetection3DLearner(
        model_config_path=config, device=device, checkpoint_after_iter=1000, return_uncertainty=True
    )
    learner.load(model_path)

    results = {}
    with open(os.path.join(model_path, "eval_" + eval_suffix + ".txt"), "w") as f:
        for samples in samples_list:
            print("samples =", samples)
            try:
                result = learner.eval(dataset, verbose=True, samples=samples)
                results[samples] = result
                f.write("samples = " + str(samples) + "\n")
                f.write("3d AP: " + str(result[2][0, :, 0]))
                f.write("\n\n")
            except Exception as e:
                f.write("samples = " + str(samples) + "\n")
                f.write("3d AP: [-1, -1, -1]")
                f.write(str(e))
                f.write("\n\n")

def test_ens_model(
    model_type,
    device="cuda:0",
    name="pointpillars_car",
    samples_list=[1],
    config="xyres_16.proto",
    eval_suffix="ens",
):

    config = os.path.join(config_roots[model_type], config,)
    model_path = os.path.join(temp_dir, name)

    results = {}
    with open(os.path.join(model_path, "eval_" + eval_suffix + ".txt"), "w") as f:
        for samples in samples_list:
            print("samples =", samples)
            learner = VoxelObjectDetection3DLearner(
                model_config_path=config, device=device, checkpoint_after_iter=1000, return_uncertainty=True,
                num_ensembles=samples
            )
            learner.load(model_path)

            result = learner.eval(dataset, verbose=True, samples=samples)
            results[samples] = result
            f.write("samples = " + str(samples) + "\n")
            f.write("3d AP: " + str(result[2][0, :, 0]))
            f.write("\n\n")


def test_pointpillars(
    device="cuda:0",
    name="pointpillars_car",
    samples_list=[1],
    config="xyres_16.proto",
    eval_suffix="classic",
):
    return test_model(
        "pointpillars",
        device=device,
        name=name,
        samples_list=samples_list,
        config=config,
        eval_suffix=eval_suffix,
    )


def test_tanet(
    device="cuda:0",
    name="tanet_car",
    samples_list=[1],
    config="xyres_16.proto",
    eval_suffix="classic",
):
    return test_model(
        "tanet",
        device=device,
        name=name,
        samples_list=samples_list,
        config=config,
        eval_suffix=eval_suffix,
    )


def test_tapp(
    device="cuda:0",
    name="tapp_car",
    samples_list=[1],
    config="tapp_16.proto",
    eval_suffix="classic",
):
    return test_model(
        "tanet",
        device=device,
        name=name,
        samples_list=samples_list,
        config=config,
        eval_suffix=eval_suffix,
    )


def test_vnn_pointpillars(
    device="cuda:0",
    name="vnn_pointpillars_car",
    samples_list=[1, 2, 3, 4, 5, 6, 7, 8, 9, 10],
    config="vnn_xyres_16.proto",
    eval_suffix="vnn",
):
    return test_pointpillars(
        device=device,
        name=name,
        samples_list=samples_list,
        config=config,
        eval_suffix=eval_suffix,
    )


def test_vnna_pointpillars(
    device="cuda:0",
    name="vnna_pointpillars_car",
    samples_list=[1, 2, 3, 4, 5, 6, 7, 8, 9, 10],
    config="vnna_xyres_16.proto",
    eval_suffix="vnn",
):
    return test_pointpillars(
        device=device,
        name=name,
        samples_list=samples_list,
        config=config,
        eval_suffix=eval_suffix,
    )


def test_vnn_tanet(
    device="cuda:0",
    name="vnn_tanet_car",
    samples_list=[1, 2, 3, 4, 5],
    config="vnn_xyres_16.proto",
    eval_suffix="vnn",
):
    return test_model(
        "tanet",
        device=device,
        name=name,
        samples_list=samples_list,
        config=config,
        eval_suffix=eval_suffix,
    )


def train_vnn_from_classical_model(
    model_type,
    device="cuda:0",
    load=0,
    name="pointpillars_car",
    classical_name="pointpillars_car",
    samples_list=[1],
    config="vnn_xyres_16.proto",
    eval_suffix="vnn",
    eval_samples_list=[1],
    **kwargs,
):

    config = os.path.join(config_roots[model_type], config,)
    model_path = os.path.join(temp_dir, name)
    classical_model_path = os.path.join(temp_dir, classical_name)

    append_samples_to_name = samples_list is not None

    if samples_list is None:
        samples_list = [1]

    for samples in samples_list:
        if append_samples_to_name:
            model_path = os.path.join(temp_dir, name + "_s" + str(samples))

        learner = VoxelObjectDetection3DLearner(
            model_config_path=config,
            device=device,
            checkpoint_after_iter=1000,
            checkpoint_load_iter=load,
            return_uncertainty=True,
            num_ensembles=samples,
            **kwargs,
        )

        if load <= 0:
            learner.load_from_clasical(classical_model_path)
            print(":::Loaded from classical")

        learner.fit(
            dataset, model_dir=model_path, verbose=False, evaluate=True, samples=samples,
            logging_path=os.path.join(model_path, "log.txt")
        )
        learner.save(model_path)

        results = {}
        with open(os.path.join(model_path, "eval_" + eval_suffix + ".txt"), "w") as f:
            for samples in eval_samples_list:
                print("samples =", samples)
                try:
                    result = learner.eval(dataset, verbose=True, samples=samples)
                    results[samples] = result
                    f.write("samples = " + str(samples) + "\n")
                    f.write("3d AP: " + str(result[2][0, :, 0]))
                    f.write("\n\n")
                except Exception as e:
                    f.write("samples = " + str(samples) + "\n")
                    f.write("3d AP: [-1, -1, -1]")
                    f.write(str(e))
                    f.write("\n\n")
        
        return results


def train_vnn_from_classical_tanet(
    device="cuda:0",
    load=0,
    name="vnn_tanet_car",
    classical_name="tanet_car",
    samples_list=[1, 2, 3, 4],
    eval_samples_list=[1, 2, 3, 4],
    config="short_vnn_xyres_16.proto",
    eval_suffix="vnn",
    **kwargs,
):
    return train_vnn_from_classical_model(
        "tanet",
        device=device,
        load=load,
        name=name,
        classical_name=classical_name,
        samples_list=samples_list,
        config=config,
        eval_suffix=eval_suffix,
        eval_samples_list=eval_samples_list,
        **kwargs,
    )



def test_ens_tanet(
    device="cuda:0",
    name="ens_tanet_car",
    samples_list=[2, 3, 4],
    config="ens_xyres_16.proto",
    eval_suffix="vnn",
):
    return test_ens_model(
        "tanet",
        device=device,
        name=name,
        samples_list=samples_list,
        config=config,
        eval_suffix=eval_suffix,
    )


def test_vnn_tapp(
    device="cuda:0",
    name="vnn_tapp_car",
    samples_list=[1, 2, 3, 4],
    config="vnn_tapp_16.proto",
    eval_suffix="vnn",
):
    return test_model(
        "tanet",
        device=device,
        name=name,
        samples_list=samples_list,
        config=config,
        eval_suffix=eval_suffix,
    )


if __name__ == "__main__":
    fire.Fire()
